@@ -1,5 +1,8 @@
 import os
-from .misc_utils import get_setting_from_context
+
+from .misc_utils import get_setting_from_context, check_true
+from urllib.parse import urlparse
+
 
 FF_ENV_DEV = 'fourfront-dev'  # Maybe not used
 FF_ENV_HOTSEAT = 'fourfront-hotseat'
@@ -230,6 +233,27 @@ def public_url_mappings(envname):
     return CGAP_PUBLIC_URLS if is_cgap_env(envname) else FF_PUBLIC_URLS
 
 
+def is_cgap_server(server, allow_localhost=False):
+    """
+    Returns True if the given string looks like a CGAP server name. Otherwise returns False.
+
+    If allow_localhost (default False) is True, then 'localhost' will be treated as a CGAP host.
+    """
+    check_true(isinstance(server, str), "Server name must be a string.", error_class=ValueError)
+    return 'cgap' in server or (allow_localhost and 'localhost' in server)
+
+
+def is_fourfront_server(server, allow_localhost=False):
+    """
+    Returns True if the given string looks like a Fourfront server name. Otherwise returns False.
+
+    If allow_localhost (default False) is True, then 'localhost' will be treated as a Fourfront host.
+    """
+    check_true(isinstance(server, str), "Server name must be a string.", error_class=ValueError)
+    return (("fourfront" in server or "4dnucleome" in server) and not is_cgap_server(server)
+            or (allow_localhost and 'localhost' in server))
+
+
 def is_cgap_env(envname):
     """
     Returns True of the given string looks like a CGAP elasticbeanstalk environment name.
@@ -357,3 +381,90 @@ def infer_foursight_from_env(request, envname):
                 return FF_STAGING_IDENTIFIER
         else:
             return envname[len('fourfront-'):]  # if not data/staging, behaves exactly like CGAP
+
+
+def full_env_name(envname):
+    """
+    Given the possibly-short name of a Fourfront or CGAP beanstalk environment, return the long name.
+
+    The short name is allowed to omit 'fourfront-' but the long name is not.
+
+    Examples:
+        full_env_name('cgapdev') => 'fourfront-cgapdev'
+        full_env_name('fourfront-cgapdev') => 'fourfront-cgapdev'
+
+    Args:
+        envname str: the short or long name of a beanstalk environment
+
+    Returns:
+        a string that is the long name of the specified beanstalk environment
+    """
+    if envname in ('data', 'staging'):
+        raise ValueError("The special token '%s' is not a beanstalk environment name." % envname)
+    elif not envname.startswith('fourfront-'):
+        return 'fourfront-' + envname
+    else:
+        return envname
+
+
+def full_cgap_env_name(envname):
+    check_true(isinstance(envname, str) and "cgap" in envname, "The envname is not a CGAP env name.",
+               error_class=ValueError)
+    return full_env_name(envname)
+
+
+def full_fourfront_env_name(envname):
+    check_true(isinstance(envname, str) and "cgap" not in envname, "The envname is not a Fourfront env name.",
+               error_class=ValueError)
+    return full_env_name(envname)
+
+
+def classify_server_url(url, raise_error=True):
+    """
+    Given a server url, returns a dictionary of information about how it relates to the Fourfront & CGAP ecosystem.
+
+    If a useful result cannot be be computed, and raise_error is True, an error is raised.
+    Otherwise, a three values are computed and returned as part of a single dictionary:
+
+    * a "kind", which is 'fourfront', 'cgap', 'localhost', or (if raise_error is False) 'unknown'.
+    * an "environment", which is the name of a fourfront or cgap environment (as appropriate) or 'unknown'.
+    * a boolean "is_stg_or_prd" that is True if the environment is a production (staging or production) environment,
+      and False otherwise.
+
+    Parameters:
+
+      url: a server url to be classified
+      raise_error(bool): whether to raise an error if the classification is unknown
+
+    Returns:
+
+      a dictionary of information containing keys "kind", "environment", and "is_stg_or_prd"
+    """
+
+    parsed = urlparse(url)
+    hostname = parsed.hostname
+    hostname1 = hostname.split('.', 1)[0]  # The part before the first dot (if any)
+
+    environment = get_bucket_env(hostname1)  # First approximation, maybe overridden below
+
+    is_stg_or_prd = is_stg_or_prd_env(hostname1)
+
+    if hostname1 == 'localhost' or hostname == '127.0.0.1':
+        environment = 'unknown'
+        kind = 'localhost'
+    elif 'cgap' in hostname1:
+        kind = 'cgap'
+    elif is_stg_or_prd or 'fourfront-' in hostname1:
+        kind = 'fourfront'
+    else:
+        if raise_error:
+            raise RuntimeError("%s is not a Fourfront or CGAP server." % url)
+        else:
+            environment = 'unknown'
+            kind = 'unknown'
+
+    return {
+        'kind': kind,
+        'environment': environment,
+        'is_stg_or_prd': is_stg_or_prd
+    }
