@@ -1,24 +1,21 @@
 from __future__ import print_function
-import sys
+
+import boto3
 import json
 import os
-import time
 import random
-import boto3
+import requests
+import time
+import urllib.parse as urlparse
+
+from collections import namedtuple
+from elasticsearch.exceptions import AuthorizationException
+from urllib.parse import urlencode
 from . import (
     s3_utils,
     es_utils,
 )
 from .misc_utils import PRINT
-import requests
-from elasticsearch.exceptions import AuthorizationException
-# urlparse import differs between py2 and 3
-if sys.version_info[0] < 3:
-    import urlparse
-    from urllib import urlencode as urlencode
-else:
-    import urllib.parse as urlparse
-    from urllib.parse import urlencode
 
 
 # TODO (C4-92, C4-102): Probably to centralize this information in env_utils. Also figure out relation to CGAP.
@@ -187,6 +184,15 @@ def purge_request_with_retries(request_fxn, url, auth, verb, **kwargs):
     return final_res
 
 
+REQUESTS_VERBS = {
+    'GET': requests.get,
+    'POST': requests.post,
+    'PATCH': requests.patch,
+    'PUT': requests.put,
+    'DELETE': requests.delete,
+}
+
+
 def authorized_request(url, auth=None, ff_env=None, verb='GET',
                        retry_fxn=standard_request_with_retries, **kwargs):
     """
@@ -205,6 +211,8 @@ def authorized_request(url, auth=None, ff_env=None, verb='GET',
     OR
     authorized_request('https://data.4dnucleome.org/<some path>', ff_env='fourfront-webprod')
     """
+    # Save to uncomment if debugging unit tests...
+    # print("authorized_request\n URL=%s\n auth=%s\n ff_env=%s\n verb=%s\n" % (url, auth, ff_env, verb))
     use_auth = unified_authentication(auth, ff_env)
     headers = kwargs.get('headers')
     if not headers:
@@ -212,21 +220,19 @@ def authorized_request(url, auth=None, ff_env=None, verb='GET',
     if 'timeout' not in kwargs:
         kwargs['timeout'] = 60  # default timeout
 
-    verbs = {'GET': requests.get,
-             'POST': requests.post,
-             'PATCH': requests.patch,
-             'PUT': requests.put,
-             'DELETE': requests.delete,
-             }
     try:
-        the_verb = verbs[verb.upper()]
+        the_verb = REQUESTS_VERBS[verb.upper()]
     except KeyError:
-        raise Exception("Provided verb %s is not valid. Must one of: %s" % (verb.upper(), ', '.join(verbs.keys())))
+        raise ValueError("Provided verb %s is not valid. Must one of: %s"
+                         % (verb.upper(), ', '.join(REQUESTS_VERBS.keys())))
     # automatically detect a search and overwrite the retry if it is standard
     if '/search/' in url and retry_fxn == standard_request_with_retries:
         retry_fxn = search_request_with_retries
     # use the given retry function. MUST TAKE THESE PARAMS!
-    return retry_fxn(the_verb, url, use_auth, verb, **kwargs)
+    result = retry_fxn(the_verb, url, use_auth, verb, **kwargs)
+    # Save to uncomment if debugging unit tests...
+    # print("authorized_request result=", json.dumps(result.json(), indent=2))
+    return result
 
 
 def get_metadata(obj_id, key=None, ff_env=None, check_queue=False, add_on=''):
@@ -558,6 +564,7 @@ def get_associated_qc_metrics(uuid, key=None, ff_env=None, include_processed_fil
     representing a quality metric.
 
     Args:
+        uuid: uuid of an experimentSet
         key: authentication key for ff_env (see get_authentication_with_server)
         ff_env: The relevant ff beanstalk environment name.
         include_processed_files: if False will exclude QC metrics on processed files
@@ -590,7 +597,6 @@ def get_associated_qc_metrics(uuid, key=None, ff_env=None, include_processed_fil
     organism = None
     experiment_type = None
     experiment_subclass = None
-    description = None
     biosource_summary = None
 
     resp = get_metadata(uuid, key=key, ff_env=ff_env)
@@ -634,8 +640,8 @@ def get_associated_qc_metrics(uuid, key=None, ff_env=None, include_processed_fil
                 result.update(exp_qc_metrics)
 
     description = resp.get('dataset_label', None)
-    ES_qc_metrics = fetch_files_qc_metrics(resp, associated_files, key=key, ff_env=ff_env)
-    if ES_qc_metrics:
+    es_qc_metrics = fetch_files_qc_metrics(resp, associated_files, key=key, ff_env=ff_env)
+    if es_qc_metrics:
         meta_info = {'experiment_description': description,
                      'organism': organism,
                      'experiment_type': experiment_type,
@@ -644,9 +650,9 @@ def get_associated_qc_metrics(uuid, key=None, ff_env=None, include_processed_fil
                      'source_experimentSet': resp['accession'],
                      'biosource_summary': biosource_summary
                      }
-        for qc_metric in ES_qc_metrics.values():
+        for qc_metric in es_qc_metrics.values():
             qc_metric.update(meta_info)
-        result.update(ES_qc_metrics)
+        result.update(es_qc_metrics)
 
     return result
 
@@ -874,7 +880,7 @@ def get_schema_names(key=None, ff_env=None):
     return schema_name
 
 
-def expand_es_metadata(uuid_list, key=None, ff_env=None, store_frame='raw', add_pc_wfr=False, ignore_field=[],
+def expand_es_metadata(uuid_list, key=None, ff_env=None, store_frame='raw', add_pc_wfr=False, ignore_field=None,
                        use_generator=False, es_client=None):
     """
     starting from list of uuids, tracks all linked items in object frame by default
@@ -905,6 +911,8 @@ def expand_es_metadata(uuid_list, key=None, ff_env=None, store_frame='raw', add_
     # TODO: if more file types (currently FileFastq and FileProcessed) get workflowrun calculated properties
             we need to add them to the add_from_embedded dictionary.
     """
+    if ignore_field is None:
+        ignore_field = []
     # assert that the used parameter is correct
     accepted_frames = ['raw', 'object', 'embedded']
     if store_frame not in accepted_frames:
@@ -919,6 +927,7 @@ def expand_es_metadata(uuid_list, key=None, ff_env=None, store_frame='raw', add_
         return my_dict
 
     auth = get_authentication_with_server(key, ff_env)
+    es_url = None
     if es_client is None:  # set up an es client if none is provided
         es_url = get_health_page(key=auth)['elasticsearch']
         es_client = es_utils.create_es_client(es_url, use_aws_auth=True)
@@ -1010,21 +1019,60 @@ def expand_es_metadata(uuid_list, key=None, ff_env=None, store_frame='raw', add_
     return store, list(item_uuids)
 
 
-def get_health_page(key=None, ff_env=None):
-    """
-    Simple function to return the json for a FF health page given keys or
-    ff_env. Will return json containing an error rather than raising an
-    exception if this fails, since this function should tolerate failure
-    """
+def _get_page(*, page, key=None, ff_env=None):
+    """ Wrapper for commonly used code to GET a page from an environment
+        Given keys or ff_env, will return json containing an error rather than raising an
+        exception if this fails, since this function should tolerate failure """
     try:
         auth = get_authentication_with_server(key, ff_env)
-        health_res = authorized_request(auth['server'] + '/health', auth=auth, verb='GET')
-        ret = get_response_json(health_res)
+        page_res = authorized_request(auth['server'] + page, auth=auth, verb='GET')
+        ret = get_response_json(page_res)
     except Exception as exc:
         ret = {'error': str(exc)}
     return ret
 
 
+def get_health_page(key=None, ff_env=None):
+    """
+    Simple function to return the json for a FF health page
+    """
+    return _get_page(page='/health', key=key, ff_env=ff_env)
+
+
+def get_counts_page(key=None, ff_env=None):
+    """ Gets DB/ES counts page in JSON """
+    return _get_page(page='/counts', key=key, ff_env=ff_env)
+
+
+def get_indexing_status(key=None, ff_env=None):
+    """ Gets indexing status counts page in JSON """
+    return _get_page(page='/indexing_status', key=key, ff_env=ff_env)
+
+
+# namedtuple definition used below and can be imported elsewhere
+CountSummary = namedtuple('CountSummary', ['are_even', 'summary_total'])
+
+
+def get_counts_summary(env):
+    """ Returns a named tuple given an FF name to check representing the counts state.
+            CountSummary
+                are_even: boolean on whether or not counts are even
+                summary_total: raw value of counts
+    """
+    totals = get_counts_page(ff_env=env)
+    if 'error' in totals:  # error encountered getting page, assume false and return error
+        return CountSummary(are_even=False, summary_total=totals)
+    totals = totals['db_es_total'].split()
+
+    # example value of split totals: ['DB:', '74048', 'ES:', '74048']
+    # or ['DB:', '887', 'ES:', '888', '<', 'ES', 'has', '1', 'more', 'items', '>']
+    db_total = int(totals[1])
+    es_total = int(totals[3])
+    if db_total > es_total or es_total > db_total:
+        return CountSummary(are_even=False, summary_total=totals)
+    return CountSummary(are_even=True, summary_total=totals)
+
+
 class SearchESMetadataHandler(object):
     """
     Wrapper class for executing lucene queries directly on ES.
@@ -1106,8 +1154,8 @@ def unified_authentication(auth=None, ff_env=None):
     elif isinstance(auth, tuple) and len(auth) == 2:
         use_auth = auth
     if not use_auth:
-        raise Exception("Must provide a valid authorization key or ff "
-                        "environment. You gave: %s (key), %s (ff_env)" % (auth, ff_env))
+        raise ValueError("Must provide a valid authorization key or ff environment."
+                         " You gave: %s (key), %s (ff_env)" % (auth, ff_env))
     return use_auth
 
 
@@ -1123,14 +1171,16 @@ def get_authentication_with_server(auth=None, ff_env=None):
     if not isinstance(auth, dict) or not {'key', 'secret', 'server'} <= set(auth.keys()):
         # must have ff_env if we want to get the key
         if not ff_env:
-            raise Exception("ERROR GETTING SERVER!\nMust provide dictionary auth with"
-                            " 'server' or ff environment. You gave: %s (auth), %s (ff_env)"
-                            % (auth, ff_env))
+            # TODO: Refactor these ValueErrors to do messaging in an error class where it doesn't clutter functionality.
+            raise ValueError("ERROR GETTING SERVER!"
+                             "\nMust provide dictionary auth with 'server' or ff environment."
+                             " You gave: %s (auth), %s (ff_env)"
+                             % (auth, ff_env))
         auth = s3_utils.s3Utils(env=ff_env).get_access_keys()
         if 'server' not in auth:
-            raise Exception("ERROR GETTING SERVER!\nAuthentication retrieved using "
-                            " ff environment does not have server information. Found: %s (auth)"
-                            ", %s (ff_env)" % (auth, ff_env))
+            raise ValueError("ERROR GETTING SERVER!"
+                             "\nAuthentication retrieved using ff environment does not have server information."
+                             " Found: %s (auth), %s (ff_env)" % (auth, ff_env))
     # ensure that the server does not end with '/'
     if auth['server'].endswith('/'):
         auth['server'] = auth['server'][:-1]
@@ -1144,9 +1194,9 @@ def stuff_in_queues(ff_env, check_secondary=False):
     If check_secondary is True, will also check the secondary queue.
     """
     if not ff_env:
-        raise Exception("Must provide a full fourfront environment name to "
-                        "this function (such as 'fourfront-webdev'). You gave: "
-                        "%s" % ff_env)
+        raise ValueError("Must provide a full fourfront environment name to "
+                         "this function (such as 'fourfront-webdev'). You gave: "
+                         "%s" % ff_env)
     stuff_in_queue = False
     client = boto3.client('sqs', region_name='us-east-1')
     queue_names = ['-indexer-queue']
@@ -1179,7 +1229,6 @@ def get_response_json(res):
     Very simple function to return json from a response or raise an error if
     it is not present. Used with the metadata functions.
     """
-    res_json = None
     try:
         res_json = res.json()
     except Exception:
@@ -1215,7 +1264,9 @@ def update_url_params_and_unparse(url, url_params):
     Takes a string url and url params (in format of what is returned by
     get_url_params). Returns a string url param with newly formatted params
     """
-    parsed_url = urlparse.urlparse(url)._replace(query=urlencode(url_params, True))
+    # Note: Although it upsets linting tools, ._replace() is an advertised interface of url.parse -kmp 17-Oct-2020
+    #       See https://docs.python.org/3/library/urllib.parse.html
+    parsed_url = urlparse.urlparse(url)._replace(query=urlencode(url_params, True))  # noQA
     return urlparse.urlunparse(parsed_url)
 
 
@@ -1224,8 +1275,6 @@ def convert_param(parameter_dict, vals_as_string=False):
     converts dictionary format {argument_name: value, argument_name: value, ...}
     to {'workflow_argument_name': argument_name, 'value': value}
     """
-    # Not needed? -kmp & Will 30-Mar-2020
-    # print(str(parameter_dict))
     metadata_parameters = []
     for k, v in parameter_dict.items():
         # we need this to be a float or integer if it really is, else a string
